#include <ctype.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <map>
using namespace std;

#include "vm.h"

char *memory;
unsigned int memorysize = 16 * 1024;
unsigned int PC;
unsigned int BP;
//SP points at the last used byte;
unsigned int SP;

/*
* Check, if it is possible to place argument
* of size argsize at address
* Throw exception, if address out of range
*/

void
check_address (UINT address, size_t argsize)
{
  if (address + argsize > memorysize)
    throw segmentation_fault (PC, address);
}

/*
* Get the reference to the argument described by aa
* Two versions: for real and integer argument
*/


REAL & get_real_operand_ref (const argaddr & aa)
{
  UINT address;
  if (aa.am == AM_INDIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      address = *((UINT *) (memory + address));
      check_address (address, sizeof (REAL));
      return *(REAL *) (memory + address);
    }
  else
  if(aa.am == AM_DIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (REAL));
      return *(REAL *) (memory + address);
    }
  else
    throw segmentation_fault(PC,0);
};

UINT & get_int_operand_ref (const argaddr & aa)
{
  UINT address;
  if (aa.am == AM_INDIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      address = *((UINT *) (memory + address));
      check_address (address, sizeof (UINT));
      return *(UINT *) (memory + address);
    }
  else
  if(aa.am == AM_DIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      return *(UINT *) (memory + address);
    }
  else
    throw segmentation_fault(PC,0);
};


/*
* Get the value of the argument described by aa
* Two versions: for real and integer argument
*/

REAL
get_real_operand (const argaddr & aa)
{
  UINT address;
  if (aa.am == AM_INDIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      address = *((UINT *) (memory + address));
      check_address (address, sizeof (REAL));
      return *(REAL *) (memory + address);
    }
  else if (aa.am == AM_DIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (REAL));
      return *(REAL *) (memory + address);
    }
  else
    return aa.realval;
};

UINT
get_int_operand (const argaddr & aa)
{
  UINT address;
  if (aa.am == AM_INDIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      address = *((UINT *) (memory + address));
      check_address (address, sizeof (UINT));
      return *(UINT *) (memory + address);
    }
  else if (aa.am == AM_DIRECT)
    {
      address = aa.intval + ((aa.br == BR_BP) ? BP : 0);
      check_address (address, sizeof (UINT));
      return *(UINT *) (memory + address);
    }
  else
    return aa.intval + ((aa.br == BR_BP) ? BP : 0);

};

/*
* Functions checking types of arguments
*/ 

void
s_threearg (const instruction & i)
{
  if ((i.args[0].am == AM_NONE) || (i.args[1].am == AM_NONE)
      || (i.args[2].am == AM_NONE))
    throw syntax_error ("Three arguments rewuired");
  if (i.args[2].am == AM_IMMEDIATE)
    throw syntax_error ("Last argument is not an lvalue");
};

void
s_threearg_jump (const instruction & i)
{
  if ((i.args[0].am == AM_NONE) || (i.args[1].am == AM_NONE)
      || (i.args[2].am == AM_NONE))
    throw syntax_error ("Three arguments required");
};

void
s_none (const instruction & i)
{
  if ((i.args[0].am != AM_NONE) || (i.args[1].am != AM_NONE)
      || (i.args[2].am != AM_NONE))
    throw syntax_error ("Opcode requires no arguments");
};

void
s_twoarg (const instruction & i)
{
  if ((i.args[0].am == AM_NONE) || (i.args[1].am == AM_NONE)
      || (i.args[2].am != AM_NONE))
    throw syntax_error ("Two arguments required");
  if (i.args[1].am == AM_IMMEDIATE)
    throw syntax_error ("Last argument is not an lvalue");
};

void
s_onearg (const instruction & i)
{
  if ((i.args[0].am == AM_NONE) || (i.args[1].am != AM_NONE)
      || (i.args[2].am != AM_NONE))
    throw syntax_error ("Exactly one argument required");
}

void
s_oneinteger (const instruction & i)
{
  if ((i.args[0].am == AM_NONE) || (i.args[1].am != AM_NONE)
      || (i.args[2].am != AM_NONE))
    throw syntax_error ("One argument required");
  if (i.at != AT_INTEGER)
    throw syntax_error ("Integer argument required");
}

/*
* Functions handling specific instructions
*/

void
h_none (const instruction &)
{
}

void
h_mov (const instruction & i)
{
  if (i.at == AT_INTEGER)
    get_int_operand_ref (i.args[1]) = get_int_operand (i.args[0]);
  else
    get_real_operand_ref (i.args[1]) = get_real_operand (i.args[0]);
  PC++;
}

void
h_add (const instruction & i)
{
  if (i.at == AT_INTEGER)
    get_int_operand_ref (i.args[2]) =
      get_int_operand (i.args[0]) + get_int_operand (i.args[1]);
  else
    get_real_operand_ref (i.args[2]) =
      get_real_operand (i.args[0]) + get_real_operand (i.args[1]);
  PC++;
}

void
h_call (const instruction & i)
{
  if (SP < sizeof (UINT))
    throw segmentation_fault (PC, SP);
  SP -= sizeof (UINT);
  PC++;
  *((UINT *) (memory + SP)) = PC;
  PC = get_int_operand (i.args[0]);
}


void
h_enter (const instruction & i)
{
// push BP
// BP := SP
// SP := SP - number
  if (SP < sizeof (UINT))
    throw segmentation_fault (PC, SP);
  SP-=sizeof(UINT);
  *((UINT*)(memory+SP))=BP;    
  BP = SP;
  UINT arg = get_int_operand (i.args[0]);
  if (SP < arg)
    throw segmentation_fault (PC, SP);
  SP = SP - arg;
  PC++;
}

void
h_leave (const instruction &)
{
// SP:=BP
// pop BP
  SP = BP;
  if (SP + sizeof (UINT) > memorysize)
    throw segmentation_fault (PC, SP);
  BP = *((UINT *) (memory + SP));
  SP += sizeof (UINT);
  PC++;
}

void
h_return (const instruction & i)
{
  if (SP + sizeof (UINT) > memorysize)
    throw segmentation_fault (PC, SP);
  PC = *((UINT *) (memory + SP));
  SP += sizeof (UINT);
}

void
h_write (const instruction & i)
{
  if (i.at == AT_INTEGER)
    cout << (int)get_int_operand (i.args[0]) << endl;
  else
    cout << get_real_operand (i.args[0]) << endl;
  PC++;
}

void
h_jump (const instruction & i)
{
  PC = get_int_operand (i.args[0]);
}


void
h_je (const instruction & i)
{
  if (i.at == AT_INTEGER)
    if (get_int_operand (i.args[0]) == get_int_operand (i.args[1]))
      PC = get_int_operand (i.args[2]);
    else
      PC++;
  else if (get_real_operand (i.args[0]) == get_real_operand (i.args[1]))
    PC = get_int_operand (i.args[2]);
  else
    PC++;
};

void
h_jge (const instruction & i)
{
  if (i.at == AT_INTEGER)
    if ((INT)get_int_operand (i.args[0]) >= (INT)get_int_operand (i.args[1]))
      PC = get_int_operand (i.args[2]);
    else
      PC++;
  else if (get_real_operand (i.args[0]) >= get_real_operand (i.args[1]))
    PC = get_int_operand (i.args[2]);
  else
    PC++;
};

void
h_push (const instruction & i)
{
  if (i.at == AT_INTEGER)
    {
      if (SP < sizeof (UINT))
	throw segmentation_fault (PC, SP);
      SP -= sizeof (UINT);
      *((UINT *) (memory + SP)) = get_int_operand (i.args[0]);
    }
  else
    {
      if (SP < sizeof (REAL))
	throw segmentation_fault (PC, SP);
      SP -= sizeof (REAL);
      *((REAL *) (memory + SP)) = get_real_operand (i.args[0]);
    }

  PC++;
};

void
h_incsp (const instruction & i)
{
SP+=get_int_operand (i.args[0]);
PC++;
};


/*
* Main program execution loop
*/

void
execute ()
{
  PC = 0;
  BP = 0;
  SP = memorysize;
  try
  {
    while (1)
      {
	if (PC >= instructions.size ())
	  throw segmentation_fault (PC, 0);
	instruction instr = instructions[PC];
	if(instr.oc==OC_EXIT)
	 break;
        if(instr.oc>=OC_FINAL)
	  throw segmentation_fault (PC, 0);	  
	dispatch_table[instr.oc](instr);
      };
  }
  catch (segmentation_fault & sf)
  {
  if(sf.data_address)
    cerr << "Segmentation fault accessing " << sf.
      data_address << " at PC=" << sf.PC_address << endl;
  else
    cerr << "Segmentation fault at PC=" << sf.PC_address << endl;
    
  };
};
