extern char *memory;
extern unsigned int memorysize;
extern unsigned int PC;
extern unsigned int BP;
//SP points at the last used byte;
extern unsigned int SP;

extern unsigned int pass;

enum opcode
{ OC_NONE, OC_MOV, OC_ADD, OC_CALL, OC_ENTER, OC_LEAVE,
  OC_RETURN, OC_WRITE, OC_JUMP, OC_JE, OC_JGE,
  OC_PUSH, OC_INCSP, OC_EXIT, OC_FINAL
};
enum argtype
{ AT_NONE, AT_INTEGER, AT_REAL };
enum addrmode
{ AM_NONE, AM_IMMEDIATE, AM_DIRECT, AM_INDIRECT };
enum basereg
{ BR_NONE, BR_BP };

typedef int INT;            // INT and UINT be of the same size
typedef unsigned int UINT;  // and both shorter than long long
typedef double REAL;
#define IMIN INT_MIN
#define IMAX INT_MAX

struct argaddr
{
  addrmode am;
  basereg br;
  union
  {
    UINT intval;
    REAL realval;
  };
};
struct instruction
{
  opcode oc;
  argtype at;
  argaddr args[3];
};

extern map < string, opcode > opcode_table;
extern vector < void (*) (const instruction &) > dispatch_table;
extern vector < void (*) (const instruction &) > syntax_table;


class segmentation_fault
{
public:
  UINT PC_address;
  UINT data_address;
    segmentation_fault (UINT PC_addr,
			UINT data_addr):PC_address (PC_addr),
    data_address (data_addr)
  {
  };
};


extern vector < instruction > instructions;
extern map < string, UINT >symtab;
extern UINT current_address;


struct syntax_error
{
string message;
syntax_error(const string & m):message(m){};
};

struct opcode_info
{
const char* name;
opcode oc;
argtype at; // default argument type
void (*oh) (const instruction &); //opcode handler
void (*sh) (const instruction &); //syntax handler
};

void execute ();
void analyze (const char *s);
void setup_opcodes ();



void s_threearg (const instruction &);
void s_threearg_jump (const instruction &);
void s_oneinteger (const instruction &);
void s_onearg (const instruction &);
void s_none (const instruction &);
void s_one (const instruction &);
void s_twoarg (const instruction &);

void
h_none (const instruction &);
void
h_mov (const instruction &);
void
h_add (const instruction & i);
void
h_call (const instruction & i);
void
h_enter (const instruction & i);
void
h_leave (const instruction &);
void
h_return (const instruction & i);
void
h_write (const instruction & i);
void
h_jump (const instruction & i);
void
h_je (const instruction & i);
void
h_jge (const instruction & i);
void
h_push (const instruction & i);
void
h_incsp (const instruction & i);
